@@ -1,4 +1,4 @@
-# Copyright 2012 Jussi Pakkanen
+# Copyright 2012-2014 Jussi Pakkanen
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,8 +15,41 @@
 import coredata
 import environment
 import dependencies
+import mlog
 import copy, os
 
+
+known_basic_kwargs = {'ui_files': True,
+                      'moc_headers' : True,
+                      'qresources' : True,
+                      'moc_sources' : True,
+                      'install' : True,
+                      'c_pch' : True,
+                      'cpp_pch' : True,
+                      'c_args' : True,
+                      'cpp_args' : True,
+                      'cs_args' : True,
+                      'link_args' : True,
+                      'link_depends': True,
+                      'link_with' : True,
+                      'include_directories': True,
+                      'dependencies' : True,
+                      'install_dir' : True,
+                      'main_class' : True,
+                      'gui_app' : True,
+                      'extra_files' : True,
+                      'install_rpath' : True,
+                      'resources' : True,
+                      'sources' : True,
+                      'objects' : True,
+                      'native' : True,
+                      }
+
+known_shlib_kwargs = known_basic_kwargs.copy()
+known_shlib_kwargs.update({'version' : True,
+                           'soversion' : True})
+
+
 class InvalidArguments(coredata.MesonException):
     pass
 
@@ -136,10 +169,25 @@ class BuildTarget():
         self.process_sourcelist(sources)
         self.process_objectlist(objects)
         self.process_kwargs(kwargs, environment)
+        self.check_unknown_kwargs(kwargs)
         if len(self.sources) == 0 and len(self.generated) == 0:
             raise InvalidArguments('Build target %s has no sources.' % name)
         self.validate_sources()
 
+    def check_unknown_kwargs(self, kwargs):
+        # Override this method in derived classes that have more
+        # keywords.
+        self.check_unknown_kwargs_int(kwargs, known_basic_kwargs)
+
+    def check_unknown_kwargs_int(self, kwargs, known_kwargs):
+        unknowns = []
+        for k in kwargs:
+            if not k in known_kwargs:
+                unknowns.append(k)
+        if len(unknowns) > 0:
+            mlog.log(mlog.bold('Warning:'), 'Unknown keyword argument(s) in target %s: %s.' %
+                     (self.name, ', '.join(unknowns)))
+
     def process_objectlist(self, objects):
         assert(isinstance(objects, list))
         for s in objects:
@@ -159,9 +207,11 @@ class BuildTarget():
             # Holder unpacking. Ugly.
             if hasattr(s, 'glist'):
                 s = s.glist
+            if hasattr(s, 'held_object'):
+                s = s.held_object
             if isinstance(s, str):
                 self.sources.append(s)
-            elif isinstance(s, GeneratedList):
+            elif isinstance(s, GeneratedList) or isinstance(s, CustomTarget):
                 self.generated.append(s)
             else:
                 raise InvalidArguments('Bad source in target %s.' % self.name)
@@ -251,6 +301,10 @@ class BuildTarget():
         if not isinstance(cpplist, list):
             cpplist = [cpplist]
         self.add_compiler_args('cpp', cpplist)
+        cslist = kwargs.get('cs_args', [])
+        if not isinstance(cslist, list):
+            cslist = [cslist]
+        self.add_compiler_args('cs', cslist)
         self.link_args = kwargs.get('link_args', [])
         if not isinstance(self.link_args, list):
             self.link_args = [self.link_args]
@@ -267,7 +321,7 @@ class BuildTarget():
             self.set_version(kwargs['version'])
         if 'soversion' in kwargs:
             self.set_soversion(kwargs['soversion'])
-        inclist = kwargs.get('include_dirs', [])
+        inclist = kwargs.get('include_directories', [])
         if not isinstance(inclist, list):
             inclist = [inclist]
         self.add_include_dirs(inclist)
@@ -302,6 +356,16 @@ class BuildTarget():
         self.install_rpath = kwargs.get('install_rpath', '')
         if not isinstance(self.install_rpath, str):
             raise InvalidArguments('Install_rpath is not a string.')
+        resources = kwargs.get('resources', [])
+        if not isinstance(resources, list):
+            resources = [resources]
+        for r in resources:
+            if not isinstance(r, str):
+                raise InvalidArguments('Resource argument is not a string.')
+            trial = os.path.join(environment.get_source_dir(), self.subdir, r)
+            if not os.path.isfile(trial):
+                raise InvalidArguments('Tried to add non-existing resource %s.' % r)
+        self.resources = resources
 
     def get_subdir(self):
         return self.subdir
@@ -313,7 +377,12 @@ class BuildTarget():
         return self.extra_args.get(language, [])
 
     def get_dependencies(self):
-        return self.link_targets
+        transitive_deps = []
+        for t in self.link_targets:
+            transitive_deps.append(t)
+            if isinstance(t, StaticLibrary):
+                transitive_deps += t.get_dependencies()
+        return transitive_deps
 
     def get_basename(self):
         return self.name
@@ -362,12 +431,15 @@ class BuildTarget():
         [self.add_external_dep(dep) for dep in args]
 
     def link(self, target):
-        if hasattr(target, 'held_object'):
-            target = target.held_object
-        if not isinstance(target, StaticLibrary) and \
-        not isinstance(target, SharedLibrary):
-            raise InvalidArguments('Link target is not library.')
-        self.link_targets.append(target)
+        if not isinstance(target, list):
+            target = [target]
+        for t in target:
+            if hasattr(t, 'held_object'):
+                t = t.held_object
+            if not isinstance(t, StaticLibrary) and \
+            not isinstance(t, SharedLibrary):
+                raise InvalidArguments('Link target is not library.')
+            self.link_targets.append(t)
 
     def set_generated(self, genlist):
         for g in genlist:
@@ -450,7 +522,7 @@ class Generator():
             if not isinstance(a, str):
                 raise InvalidArguments('A non-string object in "arguments" keyword argument.')
         self.arglist = args
-        
+
         if 'output' not in kwargs:
             raise InvalidArguments('Generator must have "output" keyword argument.')
         outputs = kwargs['output']
@@ -510,6 +582,8 @@ class Executable(BuildTarget):
         self.prefix = ''
         self.suffix = environment.get_exe_suffix()
         suffix = environment.get_exe_suffix()
+        if len(self.sources) > 0 and self.sources[0].endswith('.cs'):
+            suffix = 'exe'
         if suffix != '':
             self.filename = self.name + '.' + suffix
         else:
@@ -519,6 +593,8 @@ class Executable(BuildTarget):
 class StaticLibrary(BuildTarget):
     def __init__(self, name, subdir, is_cross, sources, objects, environment, kwargs):
         super().__init__(name, subdir, is_cross, sources, objects, environment, kwargs)
+        if len(self.sources) > 0 and self.sources[0].endswith('.cs'):
+            raise InvalidArguments('Static libraries not supported for C#.')
         self.prefix = environment.get_static_lib_prefix()
         self.suffix = environment.get_static_lib_suffix()
         self.filename = self.prefix + self.name + '.' + self.suffix
@@ -534,10 +610,17 @@ class SharedLibrary(BuildTarget):
         self.version = None
         self.soversion = None
         super().__init__(name, subdir, is_cross, sources, objects, environment, kwargs);
-        self.prefix = environment.get_shared_lib_prefix()
-        self.suffix = environment.get_shared_lib_suffix()
+        if len(self.sources) > 0 and self.sources[0].endswith('.cs'):
+            self.suffix = 'dll'
+            self.prefix = 'lib'
+        else:
+            self.prefix = environment.get_shared_lib_prefix()
+            self.suffix = environment.get_shared_lib_suffix()
         self.importsuffix = environment.get_import_lib_suffix()
 
+    def check_unknown_kwargs(self, kwargs):
+        self.check_unknown_kwargs_int(kwargs, known_shlib_kwargs)
+
     def get_shbase(self):
         return self.prefix + self.name + '.' + self.suffix
 
@@ -582,6 +665,11 @@ class SharedLibrary(BuildTarget):
         return aliases
 
 class CustomTarget:
+    known_kwargs = {'input' : True,
+                    'output' : True,
+                    'command' : True,
+                    'install' : True,
+                    'install_dir' : True}
     def __init__(self, name, subdir, kwargs):
         self.name = name
         self.subdir = subdir
@@ -589,6 +677,13 @@ class CustomTarget:
         self.process_kwargs(kwargs)
         self.extra_files = []
         self.install_rpath = ''
+        unknowns = []
+        for k in kwargs:
+            if k not in CustomTarget.known_kwargs:
+                unknowns.append(k)
+        if len(unknowns) > 0:
+            mlog.log(mlog.bold('Warning:'), 'Unknown keyword arguments in target %s: %s' %
+                     (self.name, ', '.join(unknowns)))
 
     def process_kwargs(self, kwargs):
         self.sources = kwargs.get('input', [])
@@ -597,10 +692,13 @@ class CustomTarget:
         if 'output' not in kwargs:
             raise InvalidArguments('Missing keyword argument "output".')
         self.output = kwargs['output']
-        if not(isinstance(self.output, str)):
-            raise InvalidArguments('Output argument not a string.')
-        if '/' in self.output:
-            raise InvalidArguments('Output must not contain a path segment.')
+        if not isinstance(self.output, list):
+            self.output = [self.output]
+        for i in self.output:
+            if not(isinstance(i, str)):
+                raise InvalidArguments('Output argument not a string.')
+            if '/' in i:
+                raise InvalidArguments('Output must not contain a path segment.')
         if 'command' not in kwargs:
             raise InvalidArguments('Missing keyword argument "command".')
         cmd = kwargs['command']
@@ -615,13 +713,16 @@ class CustomTarget:
             if isinstance(c, str):
                 final_cmd.append(c)
             elif isinstance(c, dependencies.ExternalProgram):
-                final_cmd.append(c.get_command())
+                final_cmd += c.get_command()
             elif isinstance(c, BuildTarget) or isinstance(c, CustomTarget):
                 self.dependencies.append(c)
                 # GIR scanner will attempt to execute this binary but
                 # it assumes that it is in path, so always give it a full path.
-                totarget = os.path.join('.', c.get_subdir(), c.get_filename())
-                final_cmd.append(totarget)
+                tmp = c.get_filename()
+                if isinstance(tmp, str):
+                    tmp =[tmp]
+                totarget = [os.path.join('.', c.get_subdir(), i) for i in tmp]
+                final_cmd += totarget
             elif isinstance(c, list):
                 # Hackety hack, only supports one level of flattening. Should really
                 # work to arbtrary depth.
@@ -695,6 +796,7 @@ class RunTarget:
 
     def get_filename(self):
         return self.name
+
 class Jar(BuildTarget):
     def __init__(self, name, subdir, is_cross, sources, objects, environment, kwargs):
         super().__init__(name, subdir, is_cross, sources, objects, environment, kwargs);
@@ -719,7 +821,7 @@ class ConfigureFile():
 
     def get_sources(self):
         return self.sources
-    
+
     def get_subdir(self):
         return self.subdir
 
