@@ -1,4 +1,4 @@
-# Copyright 2012 Jussi Pakkanen
+# Copyright 2012-2014 Jussi Pakkanen
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -35,6 +35,22 @@ class RunResult():
         self.stdout = stdout
         self.stderr = stderr
 
+def is_osx():
+    return platform.system().lower() == 'darwin'
+
+def is_linux():
+    return platform.system().lower() == 'linux'
+
+def is_windows():
+    return platform.system().lower() == 'windows'
+
+def is_debianlike():
+    try:
+        open('/etc/debian_version', 'r')
+        return True
+    except FileNotFoundError:
+        return False
+
 gnulike_buildtype_args = {'plain' : [],
                            'debug' : ['-g'],
                            'debugoptimized' : ['-O2', '-g'],
@@ -46,11 +62,20 @@ msvc_buildtype_args = {'plain' : [],
                         'debugoptimized' : ["/MD", "/Zi", "/O2", "/Ob1", "/D"],
                         'release' : ["/MD", "/O2", "/Ob2"]}
 
-gnulike_buildtype_linker_args = {'plain' : [],
-                                  'debug' : [],
-                                  'debugoptimized' : [],
-                                  'release' : ['-Wl,-O1'],
-                                  }
+gnulike_buildtype_linker_args = {}
+
+if is_osx():
+    gnulike_buildtype_linker_args.update({'plain' : [],
+                                          'debug' : [],
+                                          'debugoptimized' : [],
+                                          'release' : [],
+                                          })
+else:
+    gnulike_buildtype_linker_args.update({'plain' : [],
+                                          'debug' : [],
+                                          'debugoptimized' : [],
+                                          'release' : ['-Wl,-O1'],
+                                          })
 
 msvc_buildtype_linker_args = {'plain' : [],
                                'debug' : [],
@@ -62,6 +87,23 @@ rust_buildtype_args = {'plain' : [],
                         'debugoptimized' : ['-g', '--opt-level', '2'],
                         'release' : ['--opt-level', '3']}
 
+mono_buildtype_args = {'plain' : [],
+                       'debug' : ['-debug'],
+                       'debugoptimized': ['-debug', '-optimize+'],
+                       'release' : ['-optimize+']}
+
+def build_unix_rpath_args(build_dir, rpath_paths, install_rpath):
+        if len(rpath_paths) == 0 and len(install_rpath) == 0:
+            return []
+        paths = ':'.join([os.path.join(build_dir, p) for p in rpath_paths])
+        if len(paths) < len(install_rpath):
+            padding = 'X'*(len(install_rpath) - len(paths))
+            if len(paths) == 0:
+                paths = padding
+            else:
+                paths = paths + ':' + padding
+        return ['-Wl,-rpath,' + paths]
+
 class CCompiler():
     def __init__(self, exelist, version, is_cross, exe_wrapper=None):
         if type(exelist) == type(''):
@@ -89,7 +131,7 @@ class CCompiler():
     def get_linker_always_args(self):
         return []
 
-    def get_soname_args(self, shlib_name, path):
+    def get_soname_args(self, shlib_name, path, soversion):
         return []
 
     def split_shlib_to_parts(self, fname):
@@ -98,16 +140,7 @@ class CCompiler():
     # The default behaviour is this, override in
     # OSX and MSVC.
     def build_rpath_args(self, build_dir, rpath_paths, install_rpath):
-        if len(rpath_paths) == 0 and len(install_rpath) == 0:
-            return []
-        paths = ':'.join([os.path.join(build_dir, p) for p in rpath_paths])
-        if len(paths) < len(install_rpath):
-            padding = 'X'*(len(install_rpath) - len(paths))
-            if len(paths) == 0:
-                paths = padding
-            else:
-                paths = paths + ':' + padding
-        return ['-Wl,-rpath,' + paths]
+        return build_unix_rpath_args(build_dir, rpath_paths, install_rpath)
 
     def get_id(self):
         return self.id
@@ -126,7 +159,7 @@ class CCompiler():
 
     def get_exelist(self):
         return self.exelist[:]
-    
+
     def get_linker_exelist(self):
         return self.exelist[:]
 
@@ -148,11 +181,14 @@ class CCompiler():
     def get_coverage_link_args(self):
         return ['-lgcov']
 
+    def get_werror_args(self):
+        return ['-Werror']
+
     def get_std_exe_link_args(self):
         return []
 
-    def get_include_arg(self, path):
-        return '-I' + path
+    def get_include_args(self, path):
+        return ['-I' + path]
 
     def get_std_shared_lib_link_args(self):
         return ['-shared']
@@ -442,6 +478,149 @@ class ObjCPPCompiler(CPPCompiler):
         if pe.returncode != 0:
             raise EnvironmentException('Executables created by ObjC++ compiler %s are not runnable.' % self.name_string())
 
+class MonoCompiler():
+    def __init__(self, exelist, version):
+        if type(exelist) == type(''):
+            self.exelist = [exelist]
+        elif type(exelist) == type([]):
+            self.exelist = exelist
+        else:
+            raise TypeError('Unknown argument to Mono compiler')
+        self.version = version
+        self.language = 'cs'
+        self.default_suffix = 'cs'
+        self.id = 'mono'
+        self.monorunner = 'mono'
+
+    def get_always_args(self):
+        return []
+
+    def get_output_args(self, fname):
+        return ['-out:' + fname]
+
+    def get_linker_always_args(self):
+        return []
+
+    def get_link_args(self, fname):
+        return ['-r:' + fname]
+
+    def get_soname_args(self, shlib_name, path, soversion):
+        return []
+
+    def get_werror_args(self):
+        return ['-warnaserror']
+
+    def split_shlib_to_parts(self, fname):
+        return (None, fname)
+
+    def build_rpath_args(self, build_dir, rpath_paths, install_rpath):
+        return []
+
+    def get_id(self):
+        return self.id
+
+    def get_dependency_gen_args(self, outtarget, outfile):
+        return []
+
+    def get_language(self):
+        return self.language
+
+    def get_default_suffix(self):
+        return self.default_suffix
+
+    def get_exelist(self):
+        return self.exelist[:]
+
+    def get_linker_exelist(self):
+        return self.exelist[:]
+
+    def get_compile_only_args(self):
+        return []
+
+    def get_linker_output_args(self, outputname):
+        return []
+
+    def get_debug_args(self):
+        return ['-g']
+
+    def get_coverage_args(self):
+        return []
+
+    def get_coverage_link_args(self):
+        return []
+
+    def get_std_exe_link_args(self):
+        return []
+
+    def get_include_args(self, path):
+        return []
+
+    def get_std_shared_lib_link_args(self):
+        return []
+
+    def can_compile(self, filename):
+        suffix = filename.split('.')[-1]
+        if suffix == 'cs':
+            return True
+        return False
+
+    def get_pic_args(self):
+        return []
+
+    def name_string(self):
+        return ' '.join(self.exelist)
+
+    def get_pch_use_args(self, pch_dir, header):
+        return []
+
+    def get_pch_name(self, header_name):
+        return ''
+
+    def sanity_check(self, work_dir):
+        src = 'sanity.cs'
+        obj = 'sanity.exe'
+        source_name = os.path.join(work_dir, src)
+        ofile = open(source_name, 'w')
+        ofile.write('''public class Sanity {
+    static public void Main () {
+    }
+}
+''')
+        ofile.close()
+        pc = subprocess.Popen(self.exelist + [src], cwd=work_dir)
+        pc.wait()
+        if pc.returncode != 0:
+            raise EnvironmentException('Mono compiler %s can not compile programs.' % self.name_string())
+        cmdlist = [self.monorunner, obj]
+        pe = subprocess.Popen(cmdlist, cwd=work_dir)
+        pe.wait()
+        if pe.returncode != 0:
+            raise EnvironmentException('Executables created by Mono compiler %s are not runnable.' % self.name_string())
+
+    def needs_static_linker(self):
+        return False
+
+    def has_header(self, hname):
+        raise EnvironmentException('Mono does not support header checks.')
+
+    def compiles(self, code):
+        raise EnvironmentException('Mono does not support compile checks.')
+
+    def run(self, code):
+        raise EnvironmentException('Mono does not support run checks.')
+
+    def sizeof(self, element, prefix, env):
+        raise EnvironmentException('Mono does not support sizeof checks.')
+
+    def alignment(self, typename, env):
+        raise EnvironmentException('Mono does not support alignment checks.')
+
+    def has_function(self, funcname, prefix, env):
+        raise EnvironmentException('Mono does not support function checks.')
+
+    def get_buildtype_args(self, buildtype):
+        return mono_buildtype_args[buildtype]
+
 class JavaCompiler():
     def __init__(self, exelist, version):
         if type(exelist) == type(''):
@@ -462,9 +641,12 @@ class JavaCompiler():
     def get_linker_always_args(self):
         return []
 
-    def get_soname_args(self, shlib_name, path):
+    def get_soname_args(self, shlib_name, path, soversion):
         return []
 
+    def get_werror_args(self):
+        return ['-Werror']
+
     def split_shlib_to_parts(self, fname):
         return (None, fname)
 
@@ -512,8 +694,8 @@ class JavaCompiler():
     def get_std_exe_link_args(self):
         return []
 
-    def get_include_arg(self, path):
-        return ''
+    def get_include_args(self, path):
+        return []
 
     def get_std_shared_lib_link_args(self):
         return []
@@ -597,6 +779,9 @@ class ValaCompiler():
     def get_exelist(self):
         return self.exelist
 
+    def get_werror_args(self):
+        return ['--fatal-warnings']
+
     def get_language(self):
         return self.language
 
@@ -668,17 +853,22 @@ class RustCompiler():
 class VisualStudioCCompiler(CCompiler):
     std_warn_args = ['/W3']
     std_opt_args= ['/O2']
-    always_args = ['/nologo', '/showIncludes']
+    vs2010_always_args = ['/nologo', '/showIncludes']
+    vs2013_always_args = ['/nologo', '/showIncludes', '/FS']
 
     def __init__(self, exelist, version, is_cross, exe_wrap):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
         self.id = 'msvc'
+        if int(version.split('.')[0]) > 17:
+            self.always_args = VisualStudioCCompiler.vs2013_always_args
+        else:
+            self.always_args = VisualStudioCCompiler.vs2010_always_args
 
     def get_always_args(self):
-        return VisualStudioCCompiler.always_args
+        return self.always_args
 
     def get_std_warn_args(self):
-        return VisualStudioCCompiler.std_warn_args
+        return self.std_warn_args
 
     def get_buildtype_args(self, buildtype):
         return msvc_buildtype_args[buildtype]
@@ -730,25 +920,30 @@ class VisualStudioCCompiler(CCompiler):
         return ['/DLL']
 
     def gen_pch_args(self, header, source, pchname):
-        return ['/Yc' + header, '/Fp' + pchname]
+        objname = os.path.splitext(pchname)[0] + '.obj'
+        return (objname, ['/Yc' + header, '/Fp' + pchname, '/Fo' + objname ])
 
     def sanity_check(self, work_dir):
-        source_name = os.path.join(work_dir, 'sanitycheckc.c')
-        binary_name = os.path.join(work_dir, 'sanitycheckc')
-        ofile = open(source_name, 'w')
+        source_name = 'sanitycheckc.c'
+        binary_name = 'sanitycheckc'
+        ofile = open(os.path.join(work_dir, source_name), 'w')
         ofile.write('int main(int argc, char **argv) { return 0; }\n')
         ofile.close()
         pc = subprocess.Popen(self.exelist + [source_name, '/Fe' + binary_name],
                               stdout=subprocess.DEVNULL,
-                              stderr=subprocess.DEVNULL)
+                              stderr=subprocess.DEVNULL,
+                              cwd=work_dir)
         pc.wait()
         if pc.returncode != 0:
             raise EnvironmentException('Compiler %s can not compile programs.' % self.name_string())
-        pe = subprocess.Popen(binary_name)
+        pe = subprocess.Popen(os.path.join(work_dir, binary_name))
         pe.wait()
         if pe.returncode != 0:
             raise EnvironmentException('Executables created by C++ compiler %s are not runnable.' % self.name_string())
 
+    def build_rpath_args(self, build_dir, rpath_paths, install_rpath):
+        return []
+
 class VisualStudioCPPCompiler(VisualStudioCCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrap):
         VisualStudioCCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
@@ -762,18 +957,19 @@ class VisualStudioCPPCompiler(VisualStudioCCompiler):
         return False
 
     def sanity_check(self, work_dir):
-        source_name = os.path.join(work_dir, 'sanitycheckcpp.cpp')
-        binary_name = os.path.join(work_dir, 'sanitycheckcpp')
-        ofile = open(source_name, 'w')
+        source_name = 'sanitycheckcpp.cpp'
+        binary_name = 'sanitycheckcpp'
+        ofile = open(os.path.join(work_dir, source_name), 'w')
         ofile.write('class BreakPlainC;int main(int argc, char **argv) { return 0; }\n')
         ofile.close()
         pc = subprocess.Popen(self.exelist + [source_name, '/Fe' + binary_name],
                               stdout=subprocess.DEVNULL,
-                              stderr=subprocess.DEVNULL)
+                              stderr=subprocess.DEVNULL,
+                              cwd=work_dir)
         pc.wait()
         if pc.returncode != 0:
             raise EnvironmentException('Compiler %s can not compile programs.' % self.name_string())
-        pe = subprocess.Popen(binary_name)
+        pe = subprocess.Popen(os.path.join(work_dir, binary_name))
         pe.wait()
         if pe.returncode != 0:
             raise EnvironmentException('Executables created by C++ compiler %s are not runnable.' % self.name_string())
@@ -782,9 +978,13 @@ GCC_STANDARD = 0
 GCC_OSX = 1
 GCC_MINGW = 2
 
-def get_gcc_soname_args(gcc_type, shlib_name, path):
+def get_gcc_soname_args(gcc_type, shlib_name, path, soversion):
+    if soversion is None:
+        sostr = ''
+    else:
+        sostr = '.' + soversion
     if gcc_type == GCC_STANDARD:
-        return ['-Wl,-soname,lib%s.so' % shlib_name]
+        return ['-Wl,-soname,lib%s.so%s' % (shlib_name, sostr)]
     elif gcc_type == GCC_OSX:
         return ['-install_name', os.path.join(path, 'lib' + shlib_name + '.dylib')]
     else:
@@ -816,8 +1016,11 @@ class GnuCCompiler(CCompiler):
     def split_shlib_to_parts(self, fname):
         return (os.path.split(fname)[0], fname)
 
-    def get_soname_args(self, shlib_name, path):
-        return get_gcc_soname_args(self.gcc_type, shlib_name, path)
+    def get_soname_args(self, shlib_name, path, soversion):
+        return get_gcc_soname_args(self.gcc_type, shlib_name, path, soversion)
+
+    def can_compile(self, filename):
+        return super().can_compile(filename) or filename.split('.')[-1] == 's' # Gcc can do asm, too.
 
 class GnuObjCCompiler(ObjCCompiler):
     std_warn_args = ['-Wall', '-Winvalid-pch']
@@ -838,8 +1041,8 @@ class GnuObjCCompiler(ObjCCompiler):
     def get_pch_suffix(self):
         return 'gch'
 
-    def get_soname_args(self, shlib_name, path):
-        return get_gcc_soname_args(self.gcc_type, shlib_name, path)
+    def get_soname_args(self, shlib_name, path, soversion):
+        return get_gcc_soname_args(self.gcc_type, shlib_name, path, soversion)
 
 class GnuObjCPPCompiler(ObjCPPCompiler):
     std_warn_args = ['-Wall', '-Winvalid-pch']
@@ -861,8 +1064,8 @@ class GnuObjCPPCompiler(ObjCPPCompiler):
     def get_pch_suffix(self):
         return 'gch'
 
-    def get_soname_args(self, shlib_name, path):
-        return get_gcc_soname_args(self.gcc_type, shlib_name, path)
+    def get_soname_args(self, shlib_name, path, soversion):
+        return get_gcc_soname_args(self.gcc_type, shlib_name, path, soversion)
 
 class ClangObjCCompiler(GnuObjCCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None):
@@ -893,6 +1096,16 @@ class ClangCCompiler(CCompiler):
     def get_pch_suffix(self):
         return 'pch'
 
+    def can_compile(self, filename):
+        return super().can_compile(filename) or filename.split('.')[-1] == 's' # Clang can do asm, too.
+
+    def get_pch_use_args(self, pch_dir, header):
+        # Workaround for Clang bug http://llvm.org/bugs/show_bug.cgi?id=15136
+        # This flag is internal to Clang (or at least not documented on the man page)
+        # so it might change semantics at any time.
+        return ['-include-pch', os.path.join (pch_dir, self.get_pch_name (header))]
+
+
 class GnuCPPCompiler(CPPCompiler):
     std_warn_args = ['-Wall', '-Winvalid-pch']
     # may need to separate the latter to extra_debug_args or something
@@ -921,8 +1134,8 @@ class GnuCPPCompiler(CPPCompiler):
     def get_pch_suffix(self):
         return 'gch'
 
-    def get_soname_args(self, shlib_name, path):
-        return get_gcc_soname_args(self.gcc_type, shlib_name, path)
+    def get_soname_args(self, shlib_name, path, soversion):
+        return get_gcc_soname_args(self.gcc_type, shlib_name, path, soversion)
 
 class ClangCPPCompiler(CPPCompiler):
     std_warn_args = ['-Wall', '-Winvalid-pch']
@@ -943,6 +1156,266 @@ class ClangCPPCompiler(CPPCompiler):
     def get_pch_suffix(self):
         return 'pch'
 
+    def get_pch_use_args(self, pch_dir, header):
+        # Workaround for Clang bug http://llvm.org/bugs/show_bug.cgi?id=15136
+        # This flag is internal to Clang (or at least not documented on the man page)
+        # so it might change semantics at any time.
+        return ['-include-pch', os.path.join (pch_dir, self.get_pch_name (header))]
+
+class FortranCompiler():
+    std_warn_args = ['-Wall']
+
+    def __init__(self, exelist, version,is_cross, exe_wrapper=None):
+        super().__init__()
+        self.exelist = exelist
+        self.version = version
+        self.is_cross = is_cross
+        self.exe_wrapper = exe_wrapper
+        self.language = 'fortran'
+
+    def get_id(self):
+        return self.id
+
+    def get_exelist(self):
+        return self.exelist
+
+    def get_language(self):
+        return self.language
+
+    def needs_static_linker(self):
+        return True
+
+    def sanity_check(self, work_dir):
+        source_name = os.path.join(work_dir, 'sanitycheckf.f90')
+        binary_name = os.path.join(work_dir, 'sanitycheckf')
+        ofile = open(source_name, 'w')
+        ofile.write('''program prog
+     print *, "Fortran compilation is working."
+end program prog
+''')
+        ofile.close()
+        pc = subprocess.Popen(self.exelist + [source_name, '-o', binary_name])
+        pc.wait()
+        if pc.returncode != 0:
+            raise EnvironmentException('Compiler %s can not compile programs.' % self.name_string())
+        if self.is_cross:
+            if self.exe_wrapper is None:
+                # Can't check if the binaries run so we have to assume they do
+                return
+            cmdlist = self.exe_wrapper + [binary_name]
+        else:
+            cmdlist = [binary_name]
+        pe = subprocess.Popen(cmdlist, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
+        pe.wait()
+        if pe.returncode != 0:
+            raise EnvironmentException('Executables created by Fortran compiler %s are not runnable.' % self.name_string())
+
+    def get_always_args(self):
+        return ['-pipe']
+
+    def get_linker_always_args(self):
+        return []
+
+    def get_std_warn_args(self):
+        return FortranCompiler.std_warn_args
+
+    def get_buildtype_args(self, buildtype):
+        return gnulike_buildtype_args[buildtype]
+
+    def get_buildtype_linker_args(self, buildtype):
+        return gnulike_buildtype_linker_args[buildtype]
+
+    def split_shlib_to_parts(self, fname):
+        return (os.path.split(fname)[0], fname)
+
+    def get_soname_args(self, shlib_name, path, soversion):
+        return get_gcc_soname_args(self.gcc_type, shlib_name, path, soversion)
+
+    def get_dependency_gen_args(self, outtarget, outfile):
+        # Disabled until this is fixed:
+        # https://gcc.gnu.org/bugzilla/show_bug.cgi?id=62162
+        #return ['-cpp', '-MMD', '-MQ', outtarget]
+        return []
+
+    def get_output_args(self, target):
+        return ['-o', target]
+
+    def get_compile_only_args(self):
+        return ['-c']
+
+    def get_linker_exelist(self):
+        return self.exelist[:]
+
+    def get_linker_output_args(self, outputname):
+        return ['-o', outputname]
+
+    def can_compile(self, src):
+        suffix = os.path.splitext(src)[1].lower()
+        if suffix == '.f' or suffix == '.f95' or suffix == '.f90':
+            return True
+        return False
+
+    def get_include_args(self, path):
+        return ['-I' + path]
+
+    def get_module_outdir_args(self, path):
+        return ['-J' + path]
+
+    def get_depfile_suffix(self):
+        return 'd'
+
+    def get_std_exe_link_args(self):
+        return []
+
+    def build_rpath_args(self, build_dir, rpath_paths, install_rpath):
+        return build_unix_rpath_args(build_dir, rpath_paths, install_rpath)
+
+    def module_name_to_filename(self, module_name):
+        return module_name.lower() + '.mod'
+
+class GnuFortranCompiler(FortranCompiler):
+    def __init__(self, exelist, version, gcc_type, is_cross, exe_wrapper=None):
+        super().__init__(exelist, version, is_cross, exe_wrapper=None)
+        self.gcc_type = gcc_type
+        self.id = 'gcc'
+
+class G95FortranCompiler(FortranCompiler):
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+        super().__init__(exelist, version, is_cross, exe_wrapper=None)
+        self.id = 'g95'
+
+    def get_module_outdir_args(self, path):
+        return ['-fmod='+path]
+
+class SunFortranCompiler(FortranCompiler):
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+        super().__init__(exelist, version, is_cross, exe_wrapper=None)
+        self.id = 'sun'
+
+    def get_dependency_gen_args(self, outtarget, outfile):
+        return ['-fpp']
+
+    def get_always_args(self):
+        return []
+
+    def get_std_warn_args(self):
+        return []
+
+    def get_module_outdir_args(self, path):
+        return ['-moddir='+path]
+
+class IntelFortranCompiler(FortranCompiler):
+    std_warn_args = ['-warn', 'all']
+    
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+        super().__init__(exelist, version, is_cross, exe_wrapper=None)
+        self.id = 'intel'
+        
+    def get_module_outdir_args(self, path):
+        return ['-module', path]
+
+    def get_always_args(self):
+        return []
+
+    def can_compile(self, src):
+        suffix = os.path.splitext(src)[1].lower()
+        if suffix == '.f' or suffix == '.f90':
+            return True
+        return False
+
+    def get_std_warn_args(self):
+        return IntelFortranCompiler.std_warn_args
+
+class PathScaleFortranCompiler(FortranCompiler):
+    std_warn_args = ['-fullwarn']
+
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+        super().__init__(exelist, version, is_cross, exe_wrapper=None)
+        self.id = 'pathscale'
+
+    def get_module_outdir_args(self, path):
+        return ['-module', path]
+
+    def get_always_args(self):
+        return []
+
+    def can_compile(self, src):
+        suffix = os.path.splitext(src)[1].lower()
+        if suffix == '.f' or suffix == '.f90' or suffix == '.f95':
+            return True
+        return False
+
+    def get_std_warn_args(self):
+        return PathScaleFortranCompiler.std_warn_args
+
+class PGIFortranCompiler(FortranCompiler):
+    std_warn_args = ['-Minform=inform']
+
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+        super().__init__(exelist, version, is_cross, exe_wrapper=None)
+        self.id = 'pgi'
+
+    def get_module_outdir_args(self, path):
+        return ['-module', path]
+
+    def get_always_args(self):
+        return []
+
+    def can_compile(self, src):
+        suffix = os.path.splitext(src)[1].lower()
+        if suffix == '.f' or suffix == '.f90' or suffix == '.f95':
+            return True
+        return False
+
+    def get_std_warn_args(self):
+        return PGIFortranCompiler.std_warn_args
+
+
+class Open64FortranCompiler(FortranCompiler):
+    std_warn_args = ['-fullwarn']
+
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+        super().__init__(exelist, version, is_cross, exe_wrapper=None)
+        self.id = 'open64'
+
+    def get_module_outdir_args(self, path):
+        return ['-module', path]
+
+    def get_always_args(self):
+        return []
+
+    def can_compile(self, src):
+        suffix = os.path.splitext(src)[1].lower()
+        if suffix == '.f' or suffix == '.f90' or suffix == '.f95':
+            return True
+        return False
+
+    def get_std_warn_args(self):
+        return Open64FortranCompiler.std_warn_args
+
+class NAGFortranCompiler(FortranCompiler):
+    std_warn_args = []
+
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None):
+        super().__init__(exelist, version, is_cross, exe_wrapper=None)
+        self.id = 'nagfor'
+
+    def get_module_outdir_args(self, path):
+        return ['-mdir', path]
+
+    def get_always_args(self):
+        return []
+
+    def can_compile(self, src):
+        suffix = os.path.splitext(src)[1].lower()
+        if suffix == '.f' or suffix == '.f90' or suffix == '.f95':
+            return True
+        return False
+
+    def get_std_warn_args(self):
+        return NAGFortranCompiler.std_warn_args
+
+
 class VisualStudioLinker():
     always_args = ['/NOLOGO']
     def __init__(self, exelist):
@@ -977,6 +1450,7 @@ class ArLinker():
 
     def __init__(self, exelist):
         self.exelist = exelist
+        self.id = 'ar'
 
     def build_rpath_args(self, build_dir, rpath_paths, install_rpath):
         return []
@@ -1016,7 +1490,7 @@ def find_coverage_tools():
     gcovr_exe = 'gcovr'
     lcov_exe = 'lcov'
     genhtml_exe = 'genhtml'
-    
+
     if not exe_exists([gcovr_exe, '--version']):
         gcovr_exe = None
     if not exe_exists([lcov_exe, '--version']):
@@ -1031,22 +1505,6 @@ def find_valgrind():
         valgrind_exe = None
     return valgrind_exe
 
-def is_osx():
-    return platform.system().lower() == 'darwin'
-
-def is_linux():
-    return platform.system().lower() == 'linux'
-
-def is_windows():
-    return platform.system().lower() == 'windows'
-
-def is_debianlike():
-    try:
-        open('/etc/debian_version', 'r')
-        return True
-    except FileNotFoundError:
-        return False
-
 def detect_ninja():
     for n in ['ninja', 'ninja-build']:
         # Plain 'ninja' or 'ninja -h' yields an error
@@ -1102,6 +1560,7 @@ class Environment():
             self.cross_info = CrossBuildInfo(self.coredata.cross_file)
         else:
             self.cross_info = None
+        self.cmd_line_options = options.projectoptions
 
         # List of potential compilers.
         if is_windows():
@@ -1112,6 +1571,7 @@ class Environment():
             self.default_cpp = ['c++']
         self.default_objc = ['cc']
         self.default_objcpp = ['c++']
+        self.default_fortran = ['gfortran', 'g95', 'f95', 'f90', 'f77']
         self.default_static_linker = 'ar'
         self.vs_static_linker = 'lib'
 
@@ -1147,7 +1607,7 @@ class Environment():
 
     def get_script_dir(self):
         return os.path.dirname(self.meson_script_file)
-    
+
     def get_log_dir(self):
         return self.log_dir
 
@@ -1194,7 +1654,7 @@ class Environment():
             exe_wrap = None
         for compiler in compilers:
             try:
-                basename = os.path.basename(compiler).lower() 
+                basename = os.path.basename(compiler).lower()
                 if basename == 'cl' or basename == 'cl.exe':
                     arg = '/?'
                 else:
@@ -1225,6 +1685,66 @@ class Environment():
                 return VisualStudioCCompiler([compiler], version, is_cross, exe_wrap)
         raise EnvironmentException('Unknown compiler(s): "' + ', '.join(compilers) + '"')
 
+    def detect_fortran_compiler(self, want_cross):
+        evar = 'FC'
+        if self.is_cross_build() and want_cross:
+            compilers = [self.cross_info['fortran']]
+            is_cross = True
+            exe_wrap = self.cross_info.get('exe_wrapper', None)
+        elif evar in os.environ:
+            compilers = os.environ[evar].split()
+            is_cross = False
+            exe_wrap = None
+        else:
+            compilers = self.default_fortran
+            is_cross = False
+            exe_wrap = None
+        for compiler in compilers:
+            for arg in ['--version', '-V']:
+                try:
+                   p = subprocess.Popen([compiler] + [arg], stdout=subprocess.PIPE,
+                                     stderr=subprocess.PIPE)
+                except OSError:
+                   continue
+                (out, err) = p.communicate()
+                out = out.decode()
+                err = err.decode()
+
+                version = 'unknown version'
+                vmatch = re.search(Environment.version_regex, out)
+                if vmatch:
+                    version = vmatch.group(0)
+
+                if 'GNU Fortran' in out:
+                  return GnuFortranCompiler([compiler], version, GCC_STANDARD, is_cross, exe_wrap)
+
+                if 'G95' in out:
+                  return G95FortranCompiler([compiler], version, is_cross, exe_wrap)
+
+                if 'Sun Fortran' in err:
+                  version = 'unknown version'
+                  vmatch = re.search(Environment.version_regex, err)
+                  if vmatch:
+                      version = vmatch.group(0)
+                  return SunFortranCompiler([compiler], version, is_cross, exe_wrap)
+
+                if 'ifort (IFORT)' in out:
+                  return IntelFortranCompiler([compiler], version, is_cross, exe_wrap)
+                
+                if 'PathScale EKOPath(tm)' in err:
+                  return PathScaleFortranCompiler([compiler], version, is_cross, exe_wrap)
+
+                if 'pgf90' in out:
+                  return PGIFortranCompiler([compiler], version, is_cross, exe_wrap)
+
+                if 'Open64 Compiler Suite' in err:
+                  return Open64FortranCompiler([compiler], version, is_cross, exe_wrap)
+
+                if 'NAG Fortran' in err:
+                  return NAGFortranCompiler([compiler], version, is_cross, exe_wrap)
+
+        raise EnvironmentException('Unknown compiler(s): "' + ', '.join(compilers) + '"')
+
     def get_scratch_dir(self):
         return self.scratch_dir
 
@@ -1250,7 +1770,7 @@ class Environment():
             is_cross = False
             exe_wrap = None
         for compiler in compilers:
-            basename = os.path.basename(compiler).lower() 
+            basename = os.path.basename(compiler).lower()
             if basename == 'cl' or basename == 'cl.exe':
                 arg = '/?'
             else:
@@ -1359,6 +1879,24 @@ class Environment():
             return JavaCompiler(exelist, version)
         raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
 
+    def detect_cs_compiler(self):
+        exelist = ['mcs']
+        try:
+            p = subprocess.Popen(exelist + ['--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        except OSError:
+            raise EnvironmentException('Could not execute C# compiler "%s"' % ' '.join(exelist))
+        (out, err) = p.communicate()
+        out = out.decode()
+        err = err.decode()
+        vmatch = re.search(Environment.version_regex, out)
+        if vmatch:
+            version = vmatch.group(0)
+        else:
+            version = 'unknown version'
+        if 'Mono' in out:
+            return MonoCompiler(exelist, version)
+        raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
+
     def detect_vala_compiler(self):
         exelist = ['valac']
         try:
@@ -1451,7 +1989,7 @@ class Environment():
 
     def get_source_dir(self):
         return self.source_dir
-    
+
     def get_build_dir(self):
         return self.build_dir
 
@@ -1521,6 +2059,8 @@ def get_library_dirs():
         # I just don't know what the correct solution is.
         if plat == 'i686':
             plat = 'i386'
+        if plat.startswith('arm'):
+            plat = 'arm'
         unixdirs += glob('/usr/lib/' + plat + '*')
         if os.path.exists('/usr/lib64'):
             unixdirs.append('/usr/lib64')
@@ -1540,6 +2080,9 @@ def get_args_from_envvars(lang):
         compile_args = os.environ.get('CXXFLAGS', '').split()
         link_args = compile_args + os.environ.get('LDFLAGS', '').split()
         compile_args += os.environ.get('CPPFLAGS', '').split()
+    elif lang == 'fortran':
+        compile_args = os.environ.get('FFLAGS', '').split()
+        link_args = compile_args + os.environ.get('LDFLAGS', '').split()
     else:
         compile_args = []
         link_args = []
