@@ -1,4 +1,4 @@
-# Copyright 2012 Jussi Pakkanen
+# Copyright 2012-2014 Jussi Pakkanen
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -48,13 +48,13 @@ class TryRunResultHolder(InterpreterObject):
                              'stdout' : self.stdout_method,
                              'stderr' : self.stderr_method,
                              })
-        
+
     def returncode_method(self, args, kwargs):
         return self.res.returncode
 
     def compiled_method(self, args, kwargs):
         return self.res.compiled
-    
+
     def stdout_method(self, args, kwargs):
         return self.res.stdout
 
@@ -63,9 +63,9 @@ class TryRunResultHolder(InterpreterObject):
 
 class RunProcess(InterpreterObject):
 
-    def __init__(self, command_array, source_dir, build_dir, subdir):
+    def __init__(self, command_array, source_dir, build_dir, subdir, in_builddir=False):
         super().__init__()
-        pc = self.run_command(command_array, source_dir, build_dir, subdir)
+        pc = self.run_command(command_array, source_dir, build_dir, subdir, in_builddir)
         (stdout, stderr) = pc.communicate()
         self.returncode = pc.returncode
         self.stdout = stdout.decode().replace('\r\n', '\n')
@@ -74,13 +74,16 @@ class RunProcess(InterpreterObject):
                              'stdout' : self.stdout_method,
                              'stderr' : self.stderr_method,
                              })
-        
-    def run_command(self, command_array, source_dir, build_dir, subdir):
+
+    def run_command(self, command_array, source_dir, build_dir, subdir, in_builddir):
         cmd_name = command_array[0]
         env = {'MESON_SOURCE_ROOT' : source_dir,
                'MESON_BUILD_ROOT' : build_dir,
                'MESON_SUBDIR' : subdir}
-        cwd = os.path.join(source_dir, subdir)
+        if in_builddir:
+            cwd = os.path.join(build_dir, subdir)
+        else:
+            cwd = os.path.join(source_dir, subdir)
         child_env = os.environ.copy()
         child_env.update(env)
         try:
@@ -199,19 +202,20 @@ class ExternalLibraryHolder(InterpreterObject):
 
     def get_name(self):
         return self.el.name
-    
+
     def get_compile_args(self):
         return self.el.get_compile_args()
-    
+
     def get_link_args(self):
         return self.el.get_link_args()
-    
+
     def get_exe_args(self):
         return self.el.get_exe_args()
 
 class GeneratorHolder(InterpreterObject):
-    def __init__(self, args, kwargs):
+    def __init__(self, interpreter, args, kwargs):
         super().__init__()
+        self.interpreter = interpreter
         self.generator = build.Generator(args, kwargs)
         self.methods.update({'process' : self.process_method})
 
@@ -226,7 +230,7 @@ class GeneratorHolder(InterpreterObject):
             if not isinstance(a, str):
                 raise InvalidArguments('A non-string object in "process" arguments.')
         gl = GeneratedListHolder(self)
-        [gl.add_file(a) for a in args]
+        [gl.add_file(os.path.join(self.interpreter.subdir, a)) for a in args]
         return gl
 
 class GeneratedListHolder(InterpreterObject):
@@ -260,7 +264,7 @@ class Host(InterpreterObject):
         if self.environment.is_cross_build():
             return self.environment.cross_info.get('name')
         return platform.system().lower()
-    
+
     def is_big_endian_method(self, args, kwargs):
         return sys.byteorder != 'little'
 
@@ -285,7 +289,7 @@ class Headers(InterpreterObject):
 
     def get_subdir(self):
         return self.subdir
-    
+
     def get_sources(self):
         return self.sources
 
@@ -388,17 +392,18 @@ class RunTargetHolder(InterpreterObject):
         self.held_object = build.RunTarget(name, command, args, subdir)
 
 class Test(InterpreterObject):
-    def __init__(self, name, exe, is_parallel, cmd_args, env):
+    def __init__(self, name, exe, is_parallel, cmd_args, env, valgrind_args):
         InterpreterObject.__init__(self)
         self.name = name
         self.exe = exe
         self.is_parallel = is_parallel
         self.cmd_args = cmd_args
         self.env = env
-        
+        self.valgrind_args = valgrind_args
+
     def get_exe(self):
         return self.exe
-    
+
     def get_name(self):
         return self.name
 
@@ -488,7 +493,7 @@ class CompilerHolder(InterpreterObject):
             hadtxt = mlog.green('YES')
         else:
             hadtxt = mlog.red('NO')
-        mlog.log('Checking whether type "', mlog.bold(typename), 
+        mlog.log('Checking whether type "', mlog.bold(typename),
                  '" has member "', mlog.bold(membername), '": ', hadtxt, sep='')
         return had
 
@@ -563,6 +568,7 @@ class MesonMain(InterpreterObject):
                              'is_cross_build' : self.is_cross_build_method,
                              'has_exe_wrapper' : self.has_exe_wrapper_method,
                              'is_unity' : self.is_unity_method,
+                             'is_subproject' : self.is_subproject_method,
                              'current_source_dir' : self.current_source_dir_method,
                              'current_build_dir' : self.current_build_dir_method,
                              'set_install_script' : self.set_install_script_method,
@@ -574,7 +580,7 @@ class MesonMain(InterpreterObject):
         scriptbase = args[0]
         if not isinstance(scriptbase, str):
             raise InterpreterException('Set_install_script argument is not a string.')
-        scriptfile = os.path.join(self.interpreter.environment.source_dir, 
+        scriptfile = os.path.join(self.interpreter.environment.source_dir,
                                   self.interpreter.subdir, scriptbase)
         if not os.path.isfile(scriptfile):
             raise InterpreterException('Can not find install script %s.' % scriptbase)
@@ -626,6 +632,9 @@ class MesonMain(InterpreterObject):
     def is_unity_method(self, args, kwargs):
         return self.build.environment.coredata.unity
 
+    def is_subproject_method(self, args, kwargs):
+        return self.interpreter.subproject != ''
+
 class Interpreter():
 
     def __init__(self, build, subproject='', subdir=''):
@@ -635,7 +644,8 @@ class Interpreter():
         self.source_root = build.environment.get_source_dir()
         option_file = os.path.join(self.source_root, self.subdir, 'meson_options.txt')
         if os.path.exists(option_file):
-            oi = optinterpreter.OptionInterpreter(self.subproject)
+            oi = optinterpreter.OptionInterpreter(self.subproject,\
+                                                  self.build.environment.cmd_line_options)
             oi.process(option_file)
             self.build.environment.merge_options(oi.options)
         mesonfile = os.path.join(self.source_root, self.subdir, environment.build_filename)
@@ -667,7 +677,7 @@ class Interpreter():
         self.subproject_stack = []
 
     def build_func_dict(self):
-        self.funcs = {'project' : self.func_project, 
+        self.funcs = {'project' : self.func_project,
                       'message' : self.func_message,
                       'error' : self.func_error,
                       'executable': self.func_executable,
@@ -679,9 +689,12 @@ class Interpreter():
                       'run_target' : self.func_run_target,
                       'generator' : self.func_generator,
                       'test' : self.func_test,
+                      'install_headers' : self.func_install_headers,
                       'headers' : self.func_headers,
+                      'install_man' : self.func_install_man,
                       'man' : self.func_man,
                       'subdir' : self.func_subdir,
+                      'install_data' : self.func_install_data,
                       'data' : self.func_data,
                       'configure_file' : self.func_configure_file,
                       'include_directories' : self.func_include_directories,
@@ -694,7 +707,6 @@ class Interpreter():
                       'option' : self.func_option,
                       'get_option' : self.func_get_option,
                       'subproject' : self.func_subproject,
-                      'is_subproject' : self.func_is_subproject,
                       'pkgconfig_gen' : self.func_pkgconfig_gen,
                       }
 
@@ -732,9 +744,10 @@ class Interpreter():
             try:
                 self.evaluate_statement(cur)
             except Exception as e:
-                e.lineno = cur.lineno
-                e.colno = cur.colno
-                e.file = os.path.join(self.subdir, 'meson.build')
+                if not(hasattr(e, 'lineno')):
+                    e.lineno = cur.lineno
+                    e.colno = cur.colno
+                    e.file = os.path.join(self.subdir, 'meson.build')
                 raise e
             i += 1 # In THE FUTURE jump over blocks and stuff.
 
@@ -791,7 +804,7 @@ class Interpreter():
             if wanted != None:
                 if not isinstance(actual, wanted):
                     raise InvalidArguments('Incorrect argument type.')
-                
+
     def func_run_command(self, node, args, kwargs):
         if len(args) < 1:
             raise InterpreterException('Not enough arguments')
@@ -804,8 +817,14 @@ class Interpreter():
         for i in cargs:
             if not isinstance(i, str):
                 raise InterpreterException('Run_command arguments must be strings.')
-        args = [cmd] + cargs
-        return RunProcess(args, self.environment.source_dir, self.environment.build_dir, self.subdir)
+        if not isinstance(cmd, list):
+            cmd = [cmd]
+        args = cmd + cargs
+        in_builddir = kwargs.get('in_builddir', False)
+        if not isinstance(in_builddir, bool):
+            raise InterpreterException('in_builddir must be boolean.')
+        return RunProcess(args, self.environment.source_dir, self.environment.build_dir,
+                          self.subdir, in_builddir)
 
     def func_gettext(self, nodes, args, kwargs):
         if len(args) != 1:
@@ -854,9 +873,6 @@ class Interpreter():
         p = build.PkgConfigGenerator(libs, subdirs, name, description, version, filebase)
         self.build.pkgconfig_gens.append(p)
 
-    def func_is_subproject(self, nodes, args, kwargs):
-        return self.subproject != ''
-
     def func_subproject(self, nodes, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Subproject takes exactly one argument')
@@ -963,6 +979,10 @@ class Interpreter():
                     comp = self.environment.detect_java_compiler()
                     if is_cross:
                         cross_comp = comp # Java is platform independent.
+                elif lang == 'cs':
+                    comp = self.environment.detect_cs_compiler()
+                    if is_cross:
+                        cross_comp = comp # C# is platform independent.
                 elif lang == 'vala':
                     comp = self.environment.detect_vala_compiler()
                     if is_cross:
@@ -971,6 +991,10 @@ class Interpreter():
                     comp = self.environment.detect_rust_compiler()
                     if is_cross:
                         cross_comp = comp # FIXME, probably not correct.
+                elif lang == 'fortran':
+                    comp = self.environment.detect_fortran_compiler(False)
+                    if is_cross:
+                        cross_comp = self.environment.detect_fortran_compiler(True)
                 else:
                     raise InvalidCode('Tried to use unknown language "%s".' % lang)
                 comp.sanity_check(self.environment.get_scratch_dir())
@@ -1081,7 +1105,7 @@ class Interpreter():
         return tg
 
     def func_generator(self, node, args, kwargs):
-        gen = GeneratorHolder(args, kwargs)
+        gen = GeneratorHolder(self, args, kwargs)
         self.generators.append(gen)
         return gen
 
@@ -1114,26 +1138,41 @@ class Interpreter():
             if ' ' in k:
                 raise InterpreterException('Env var key must not have spaces in it.')
             env[k] = val
-        t = Test(args[0], args[1].held_object, par, cmd_args, env)
+        valgrind_args = kwargs.get('valgrind_args', [])
+        if not isinstance(valgrind_args, list):
+            valgrind_args = [valgrind_args]
+        for a in valgrind_args:
+            if not isinstance(a, str):
+                raise InterpreterException('Valgrind_arg not a string.')
+        t = Test(args[0], args[1].held_object, par, cmd_args, env, valgrind_args)
         self.build.tests.append(t)
         mlog.debug('Adding test "', mlog.bold(args[0]), '".', sep='')
 
     def func_headers(self, node, args, kwargs):
+        mlog.log(mlog.bold('Warning! '), 'Function headers() is deprecated, please use install_headers() instead.')
+        return self.func_install_headers(node, args, kwargs)
+
+    def func_install_headers(self, node, args, kwargs):
+        args = self.flatten(args)
         for a in args:
             if not isinstance(a, str):
                 raise InvalidArguments('Argument %s is not a string.' % str(a))
         h = Headers(args, kwargs)
         self.build.headers.append(h)
         return h
-    
+
     def func_man(self, node, args, kwargs):
+        mlog.log(mlog.bold('Warning! '), 'Function man() is deprecated, please use install_man() instead.')
+        return self.func_install_man(node, args, kwargs)
+
+    def func_install_man(self, node, args, kwargs):
         for a in args:
             if not isinstance(a, str):
                 raise InvalidArguments('Argument %s is not a string.' % str(a))
         m = Man(args, kwargs)
         self.build.man.append(m)
         return m
-    
+
     def func_subdir(self, node, args, kwargs):
         if len(kwargs) > 0:
             raise InvalidArguments('subdir command takes no keyword arguments.')
@@ -1169,6 +1208,10 @@ class Interpreter():
         self.subdir = prev_subdir
 
     def func_data(self, node, args, kwargs):
+        mlog.log(mlog.bold('Warning! '), 'Function data() is deprecated, please use install_data() instead.')
+        return self.func_install_data(node, args, kwargs)
+
+    def func_install_data(self, node, args, kwargs):
         if len(args ) < 1:
             raise InvalidArguments('Data function must have at least one argument: the subdirectory.')
         for a in args:
@@ -1177,7 +1220,7 @@ class Interpreter():
         data = Data(args[0], args[1:], kwargs)
         self.build.data.append(data)
         return data
-    
+
     def func_configure_file(self, node, args, kwargs):
         if len(args) > 0:
             raise InterpreterException("configure_file takes only keyword arguments.")
@@ -1313,7 +1356,7 @@ class Interpreter():
             isinstance(value, list):
             return True
         return False
-    
+
     def assignment(self, node):
         assert(isinstance(node, mparser.AssignmentNode))
         var_name = node.var_name
@@ -1413,11 +1456,34 @@ class Interpreter():
             obj = obj.get_value()
         if isinstance(obj, str):
             return self.string_method_call(obj, method_name, args)
+        if isinstance(obj, list):
+            return self.array_method_call(obj, method_name, self.reduce_arguments(args)[0])
         if not isinstance(obj, InterpreterObject):
             raise InvalidArguments('Variable "%s" is not callable.' % object_name)
         (args, kwargs) = self.reduce_arguments(args)
         return obj.method_call(method_name, args, kwargs)
-    
+
+    def array_method_call(self, obj, method_name, args):
+        if method_name == 'contains':
+            return self.check_contains(obj, args)
+        raise InterpreterException('Arrays do not have a method called "%s".' % method_name)
+
+    def check_contains(self, obj, args):
+        if len(args) != 1:
+            raise InterpreterException('Contains method takes exactly one argument.')
+        item = args[0]
+        for element in obj:
+            if isinstance(element, list):
+                found = self.check_contains(element, args)
+                if found:
+                    return True
+            try:
+                if element == item:
+                    return True
+            except Exception:
+                pass
+        return False
+
     def evaluate_if(self, node):
         assert(isinstance(node, mparser.IfClauseNode))
         for i in node.ifs:
@@ -1486,7 +1552,7 @@ class Interpreter():
         if not isinstance(r, bool):
             raise InterpreterException('Second argument to "or" is not a boolean.')
         return r
-    
+
     def evaluate_notstatement(self, cur):
         v = self.evaluate_statement(cur.value)
         if isinstance(v, mparser.BooleanNode):
