@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-# Copyright 2013 Jussi Pakkanen
+# Copyright 2013-2014 Jussi Pakkanen
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,6 +17,7 @@
 import sys, os, subprocess, time, datetime, pickle, multiprocessing, json
 import concurrent.futures as conc
 from optparse import OptionParser
+import platform
 
 tests_failed = False
 
@@ -26,19 +27,22 @@ parser.add_option('--wrapper', default=None, dest='wrapper',
 parser.add_option('--wd', default=None, dest='wd',
                   help='directory to cd into before running')
 class TestRun():
-    def __init__(self, res, returncode, duration, stdo, stde):
+    def __init__(self, res, returncode, duration, stdo, stde, cmd):
         self.res = res
         self.returncode = returncode
         self.duration = duration
         self.stdo = stdo
         self.stde = stde
+        self.cmd = cmd
 
-def write_log(logfile, test_name, result_str, stdo, stde):
+def write_log(logfile, test_name, result_str, result):
     logfile.write(result_str + '\n\n')
-    logfile.write('--- "%s" stdout ---\n' % test_name)
-    logfile.write(stdo)
+    logfile.write('--- command ---\n')
+    logfile.write(' '.join(result.cmd))
+    logfile.write('\n--- "%s" stdout ---\n' % test_name)
+    logfile.write(result.stdo)
     logfile.write('\n--- "%s" stderr ---\n' % test_name)
-    logfile.write(stde)
+    logfile.write(result.stde)
     logfile.write('\n-------\n\n')
 
 def write_json_log(jsonlogfile, test_name, result):
@@ -47,23 +51,33 @@ def write_json_log(jsonlogfile, test_name, result):
               'stderr' : result.stde,
               'result' : result.res,
               'duration' : result.duration,
-              'returncode' : result.returncode}
+              'returncode' : result.returncode,
+              'command' : result.cmd}
     jsonlogfile.write(json.dumps(result) + '\n')
 
+def run_with_mono(fname):
+    if fname.endswith('.exe') and not platform.system().lower() == 'windows':
+        return True
+    return False
+
 def run_single_test(wrap, test):
     global tests_failed
     if test.fname.endswith('.jar'):
         cmd = ['java', '-jar', test.fname]
+    elif run_with_mono(test.fname):
+        cmd = ['mono', test.fname]
     else:
         if test.is_cross:
             if test.exe_runner is None:
-                # Can not run test on cross compiled executable 
+                # Can not run test on cross compiled executable
                 # because there is no execute wrapper.
                 cmd = None
             else:
                 cmd = [test.exe_runner, test.fname]
         else:
             cmd = [test.fname]
+    if len(wrap) > 0 and 'valgrind' in wrap[0]:
+        wrap += test.valgrind_args
     if cmd is None:
         res = 'SKIP'
         duration = 0.0
@@ -88,7 +102,7 @@ def run_single_test(wrap, test):
             res = 'FAIL'
             tests_failed = True
         returncode = p.returncode
-    return TestRun(res, returncode, duration, stdo, stde)
+    return TestRun(res, returncode, duration, stdo, stde, cmd)
 
 def print_stats(numlen, tests, name, result, i, logfile, jsonlogfile):
     startpad = ' '*(numlen - len('%d' % (i+1)))
@@ -98,7 +112,7 @@ def print_stats(numlen, tests, name, result, i, logfile, jsonlogfile):
     result_str = '%s %s%s%s%s(%5.2f s)' % \
         (num, name, padding1, result.res, padding2, result.duration)
     print(result_str)
-    write_log(logfile, name, result_str, result.stdo, result.stde)
+    write_log(logfile, name, result_str, result)
     write_json_log(jsonlogfile, name, result)
 
 def drain_futures(futures):
