@@ -20,7 +20,7 @@ import dependencies
 from meson_install import InstallData
 from build import InvalidArguments
 from coredata import MesonException
-import os, sys, shutil, pickle
+import os, sys, shutil, pickle, re
 
 if environment.is_windows():
     quote_char = '"'
@@ -32,6 +32,15 @@ else:
 def ninja_quote(text):
     return text.replace(' ', '$ ').replace(':', '$:')
 
+class RawFilename():
+    def __init__(self, fname):
+        self.fname = fname
+
+    def split(self, c):
+        return self.fname.split(c)
+
+    def startswith(self, s):
+        return self.fname.startswith(s)
 
 class NinjaBuildElement():
     def __init__(self, outfilenames, rule, infilenames):
@@ -67,7 +76,9 @@ class NinjaBuildElement():
         self.elems.append((name, elems))
 
     def write(self, outfile):
-        line = 'build %s: %s %s' % (' '.join([ninja_quote(i) for i in  self.outfilenames]),\
+        # Sort inputs so the command line does not change.
+        # This allows Ninja to skip unnecessary rebuilds.
+        line = 'build %s: %s %s' % (' '.join([ninja_quote(i) for i in self.outfilenames]),\
                                     self.rule,
                                     ' '.join([ninja_quote(i) for i in self.infilenames]))
         if len(self.deps) > 0:
@@ -103,6 +114,7 @@ class NinjaBackend(backends.Backend):
         super().__init__(build, interp)
         self.source_suffix_in_objs = True
         self.ninja_filename = 'build.ninja'
+        self.fortran_deps = {}
 
     def generate(self):
         outfilename = os.path.join(self.environment.get_build_dir(), self.ninja_filename)
@@ -134,22 +146,153 @@ class NinjaBackend(backends.Backend):
         outfile.close()
         os.replace(tempfilename, outfilename)
 
+    # Get all generated headers. Any source file might need them so
+    # we need to add an order dependency to them.
+    def get_generated_headers(self, target):
+        header_deps = []
+        for gensource in target.get_generated_sources():
+            if isinstance(gensource, build.CustomTarget):
+                continue
+            for src in gensource.get_outfilelist():
+                if self.environment.is_header(src):
+                    header_deps.append(src)
+        return header_deps
+
+    def generate_target(self, target, outfile):
+        if isinstance(target, build.CustomTarget):
+            self.generate_custom_target(target, outfile)
+        if isinstance(target, build.RunTarget):
+            self.generate_run_target(target, outfile)
+        name = target.get_basename()
+        gen_src_deps = []
+        if name in self.processed_targets:
+            return
+        if isinstance(target, build.Jar):
+            self.generate_jar_target(target, outfile)
+            return
+        if 'rust' in self.environment.coredata.compilers.keys() and self.has_rust(target):
+            self.generate_rust_target(target, outfile)
+            return
+        if 'cs' in self.environment.coredata.compilers.keys() and self.has_cs(target):
+            self.generate_cs_target(target, outfile)
+            return
+        if 'vala' in self.environment.coredata.compilers.keys() and self.has_vala(target):
+            gen_src_deps += self.generate_vala_compile(target, outfile)
+        self.scan_fortran_module_outputs(target)
+        # The following deals with C/C++ compilation.
+        (gen_src, gen_other_deps) = self.process_dep_gens(outfile, target)
+        gen_src_deps += gen_src
+        self.process_target_dependencies(target, outfile)
+        self.generate_custom_generator_rules(target, outfile)
+        outname = self.get_target_filename(target)
+        obj_list = []
+        use_pch = self.environment.coredata.use_pch
+        is_unity = self.environment.coredata.unity
+        if use_pch and target.has_pch():
+            pch_objects = self.generate_pch(target, outfile)
+        else:
+            pch_objects = []
+        header_deps = gen_other_deps
+        unity_src = []
+        unity_deps = [] # Generated sources that must be built before compiling a Unity target.
+        header_deps += self.get_generated_headers(target)
+        for gensource in target.get_generated_sources():
+            if isinstance(gensource, build.CustomTarget):
+                for src in gensource.output:
+                    src = os.path.join(gensource.subdir, src)
+                    if self.environment.is_source(src) and not self.environment.is_header(src):
+                        if is_unity:
+                            unity_deps.append(os.path.join(self.environment.get_build_dir(), RawFilename(src)))
+                        else:
+                            obj_list.append(self.generate_single_compile(target, outfile, RawFilename(src), True,
+                                                                         header_deps))
+                    else:
+                        # Assume anything not specifically a source file is a header. This is because
+                        # people generate files with weird suffixes (.inc, .fh) that they then include
+                        # in their source files.
+                        header_deps.append(RawFilename(src))
+            else:
+                for src in gensource.get_outfilelist():
+                    if self.environment.is_object(src):
+                        obj_list.append(os.path.join(self.get_target_dir(target), target.get_basename() + '.dir', src))
+                    elif not self.environment.is_header(src):
+                        if is_unity:
+                            if '/' in src:
+                                rel_src = src
+                            else:
+                                rel_src = os.path.join(self.get_target_private_dir(target), src)
+                            unity_deps.append(rel_src)
+                            abs_src = os.path.join(self.environment.get_build_dir(), rel_src)
+                            unity_src.append(abs_src)
+                        else:
+                            obj_list.append(self.generate_single_compile(target, outfile, src, True,
+                                                                         header_deps=header_deps))
+        src_list = []
+        for src in gen_src_deps:
+                src_list.append(src)
+                if is_unity:
+                    unity_src.append(os.path.join(self.environment.get_build_dir(), src))
+                    header_deps.append(src)
+                else:
+                    # Generated targets are ordered deps because the must exist
+                    # before the sources compiling them are used. After the first
+                    # compile we get precise dependency info from dep files.
+                    # This should work in all cases. If it does not, then just
+                    # move them from orderdeps to proper deps.
+                    obj_list.append(self.generate_single_compile(target, outfile, src, True, [], header_deps))
+        for src in target.get_sources():
+            if src.endswith('.vala'):
+                continue
+            if not self.environment.is_header(src):
+                src_list.append(src)
+                if is_unity:
+                    abs_src = os.path.join(self.environment.get_source_dir(),
+                                           target.get_subdir(), src)
+                    unity_src.append(abs_src)
+                else:
+                    obj_list.append(self.generate_single_compile(target, outfile, src, False, [], header_deps))
+        obj_list += self.flatten_object_list(target)
+        if is_unity:
+            for src in self.generate_unity_files(target, unity_src):
+                obj_list.append(self.generate_single_compile(target, outfile, src, True, unity_deps + header_deps))
+        linker = self.determine_linker(target, src_list)
+        # Sort object list to preserve command line over multiple invocations.
+        elem = self.generate_link(target, outfile, outname, sorted(obj_list), linker, pch_objects)
+        self.generate_shlib_aliases(target, self.get_target_dir(target), outfile, elem)
+        self.processed_targets[name] = True
+
+    def process_target_dependencies(self, target, outfile):
+        for t in target.get_dependencies():
+            tname = t.get_basename()
+            if not tname in self.processed_targets:
+                self.generate_target(t, outfile)
+
+    def hackety_hack(self, hack):
+        if isinstance(hack, list):
+            return hack[0]
+        return hack
+
     def generate_custom_target(self, target, outfile):
-        ofilename = os.path.join(target.subdir, target.output)
-        deps = [os.path.join(i.get_subdir(), i.get_filename()) for i in target.get_dependencies()]
-        srcs = [os.path.join(self.build_to_src, i) for i in  target.sources]
+        ofilenames = [os.path.join(target.subdir, i) for i in target.output]
+        # FIXME, should not grab element at zero but rather expand all.
+        deps = [os.path.join(i.get_subdir(), self.hackety_hack(i.get_filename())) for i in target.get_dependencies()]
+        srcs = [os.path.join(self.build_to_src, target.subdir, i) for i in target.sources]
         deps +=  srcs
-        elem = NinjaBuildElement(ofilename, 'CUSTOM_COMMAND', deps)
+        elem = NinjaBuildElement(ofilenames, 'CUSTOM_COMMAND', deps)
         cmd = []
         for i in target.command:
+            for (j, src) in enumerate(srcs):
+                i = i.replace('@INPUT%d@' % j, src)
+            for (j, res) in enumerate(ofilenames):
+                i = i.replace('@OUTPUT%d@' % j, res)
             if i == '@INPUT@':
                 cmd += srcs
             elif i == '@OUTPUT@':
-                cmd.append(ofilename)
+                cmd += ofilenames
             else:
                 cmd.append(i)
         elem.add_item('COMMAND', cmd)
-        elem.add_item('description', 'Generating %s with a custom command.' % ofilename)
+        elem.add_item('description', 'Generating %s with a custom command.' % target.name)
         elem.write(outfile)
         self.processed_targets[target.name] = True
 
@@ -279,7 +422,7 @@ class NinjaBackend(backends.Backend):
 
     def generate_custom_install_script(self, d):
         d.install_script = self.build.install_script
-    
+
 
     def generate_header_install(self, d):
         incroot = self.environment.get_includedir()
@@ -433,6 +576,61 @@ class NinjaBackend(backends.Backend):
         elem.add_item('ARGS', commands)
         elem.write(outfile)
 
+    def generate_cs_resource_tasks(self, target, outfile):
+        args = []
+        deps = []
+        for r in target.resources:
+            rel_sourcefile = os.path.join(self.build_to_src, target.subdir, r)
+            if r.endswith('.resources'):
+                a = '-resource:' + rel_sourcefile
+            elif r.endswith('.txt') or r.endswith('.resx'):
+                ofilebase = os.path.splitext(os.path.basename(r))[0] + '.resources'
+                ofilename = os.path.join(self.get_target_dir(target), target.get_basename() + '.dir', ofilebase)
+                elem = NinjaBuildElement(ofilename, "CUSTOM_COMMAND", rel_sourcefile)
+                elem.add_item('COMMAND', ['resgen', rel_sourcefile, ofilename])
+                elem.add_item('DESC', 'Compiling resource %s.' % rel_sourcefile)
+                elem.write(outfile)
+                deps.append(ofilename)
+                a = '-resource:' + ofilename
+            else:
+                raise InvalidArguments('Unknown resource file %s.' % r)
+            args.append(a)
+        return (args, deps)
+
+    def generate_cs_target(self, target, outfile):
+        buildtype = self.environment.coredata.buildtype
+        fname = target.get_filename()
+        subdir = target.get_subdir()
+        outname_rel = os.path.join(subdir, fname)
+        src_list = target.get_sources()
+        compiler = self.get_compiler_for_source(src_list[0])
+        assert(compiler.get_language() == 'cs')
+        rel_srcs = [os.path.join(self.build_to_src, s) for s in src_list]
+        deps = []
+        commands = target.extra_args.get('cs', [])
+        commands += compiler.get_buildtype_args(buildtype)
+        if isinstance(target, build.Executable):
+            commands.append('-target:exe')
+        elif isinstance(target, build.SharedLibrary):
+            commands.append('-target:library')
+        else:
+            raise MesonException('Unknown C# target type.')
+        (resource_args, resource_deps) = self.generate_cs_resource_tasks(target, outfile)
+        commands += resource_args
+        deps += resource_deps
+        commands += compiler.get_output_args(outname_rel)
+        for l in target.link_targets:
+            commands += compiler.get_link_args(l.get_filename())
+            deps.append(l.get_filename())
+        if '-g' in commands:
+            outputs = [outname_rel, outname_rel + '.mdb']
+        else:
+            outputs = [outname_rel]
+        elem = NinjaBuildElement(outputs, 'cs_COMPILER', rel_srcs)
+        elem.add_dep(deps)
+        elem.add_item('ARGS', commands)
+        elem.write(outfile)
+
     def generate_single_java_compile(self, subdir, src, target, compiler, outfile):
         buildtype = self.environment.coredata.buildtype
         args = []
@@ -493,6 +691,8 @@ class NinjaBackend(backends.Backend):
             relsc = os.path.join(self.get_target_dir(target), target.get_basename() + '.dir', sc)
             rel_s = os.path.join(self.build_to_src, s)
             args += ['--deps', relsc + '.d']
+            if self.environment.coredata.werror:
+                args += valac.get_werror_args()
             for d in target.external_deps:
                 if isinstance(d, dependencies.PkgConfigDependency):
                     args += ['--pkg', d.name]
@@ -567,7 +767,8 @@ class NinjaBackend(backends.Backend):
         for (complist, is_cross) in ctypes:
             for compiler in complist:
                 langname = compiler.get_language()
-                if langname == 'java' or langname == 'vala' or langname == 'rust':
+                if langname == 'java' or langname == 'vala' or\
+                 langname == 'rust' or langname == 'cs':
                     continue
                 crstr = ''
                 if is_cross:
@@ -606,6 +807,16 @@ class NinjaBackend(backends.Backend):
         outfile.write(description)
         outfile.write('\n')
 
+    def generate_cs_compile_rule(self, compiler, outfile):
+        rule = 'rule %s_COMPILER\n' % compiler.get_language()
+        invoc = ' '.join([ninja_quote(i) for i in compiler.get_exelist()])
+        command = ' command = %s $ARGS $in\n' % invoc
+        description = ' description = Compiling cs target $out.\n'
+        outfile.write(rule)
+        outfile.write(command)
+        outfile.write(description)
+        outfile.write('\n')
+
     def generate_vala_compile_rules(self, compiler, outfile):
         rule = 'rule %s_COMPILER\n' % compiler.get_language()
         invoc = ' '.join([ninja_quote(i) for i in compiler.get_exelist()])
@@ -631,7 +842,7 @@ class NinjaBackend(backends.Backend):
                          invoc)
         description = ' description = Compiling Rust source $in.\n'
         depfile = ' depfile = $out.d\n'
-        
+
         depstyle = ' deps = gcc\n'
         outfile.write(rule)
         outfile.write(command)
@@ -640,11 +851,31 @@ class NinjaBackend(backends.Backend):
         outfile.write(depstyle)
         outfile.write('\n')
 
+    def generate_fortran_dep_hack(self, outfile):
+        if environment.is_windows():
+            cmd = 'cmd /C ""'
+        else:
+            cmd = 'true'
+        template = '''# Workaround for these issues:
+# https://groups.google.com/forum/#!topic/ninja-build/j-2RfBIOd_8
+# https://gcc.gnu.org/bugzilla/show_bug.cgi?id=47485
+rule FORTRAN_DEP_HACK
+ command = %s
+ description = Dep hack
+ restat = 1
+
+'''
+        outfile.write(template % cmd)
+
     def generate_compile_rule_for(self, langname, compiler, qstr, is_cross, outfile):
         if langname == 'java':
             if not is_cross:
                 self.generate_java_compile_rule(compiler, outfile)
             return
+        if langname == 'cs':
+            if not is_cross:
+                self.generate_cs_compile_rule(compiler, outfile)
+            return
         if langname == 'vala':
             if not is_cross:
                 self.generate_vala_compile_rules(compiler, outfile)
@@ -653,6 +884,8 @@ class NinjaBackend(backends.Backend):
             if not is_cross:
                 self.generate_rust_compile_rules(compiler, outfile)
             return
+        if langname == 'fortran':
+            self.generate_fortran_dep_hack(outfile)
         if is_cross:
             crstr = '_CROSS'
         else:
@@ -719,8 +952,23 @@ class NinjaBackend(backends.Backend):
                 self.generate_pch_rule_for(langname, compiler, qstr, True, outfile)
         outfile.write('\n')
 
+    def replace_outputs(self, args, private_dir, output_list):
+        newargs = []
+        regex = re.compile('@OUTPUT(\d+)@')
+        for arg in args:
+            m = regex.search(arg)
+            while m is not None:
+                index = int(m.group(1))
+                src = '@OUTPUT%d@' % index
+                arg = arg.replace(src, os.path.join(private_dir, output_list[index]))
+                m = regex.search(arg)
+            newargs.append(arg)
+        return newargs
+
     def generate_custom_generator_rules(self, target, outfile):
         for genlist in target.get_generated_sources():
+            if isinstance(genlist, build.CustomTarget):
+                continue # Customtarget has already written its output rules
             generator = genlist.get_generator()
             exe = generator.get_exe()
             if self.environment.is_cross_build() and \
@@ -733,13 +981,13 @@ class NinjaBackend(backends.Backend):
             infilelist = genlist.get_infilelist()
             outfilelist = genlist.get_outfilelist()
             if isinstance(exe, build.BuildTarget):
-                exe_file = os.path.join(self.environment.get_build_dir(), self.get_target_filename(exe))
+                exe_arr = [os.path.join(self.environment.get_build_dir(), self.get_target_filename(exe))]
             else:
-                exe_file = exe.get_command()
+                exe_arr = exe.get_command()
             base_args = generator.get_arglist()
             for i in range(len(infilelist)):
-                if len(infilelist) == len(outfilelist):
-                    sole_output = os.path.join(self.get_target_private_dir(target), outfilelist[i])
+                if len(generator.outputs) == 1:
+                    sole_output = os.path.join(self.get_target_private_dir(target), outfilelist[0])
                 else:
                     sole_output = ''
                 curfile = infilelist[i]
@@ -748,9 +996,10 @@ class NinjaBackend(backends.Backend):
                 outfiles = [os.path.join(self.get_target_private_dir(target), of) for of in outfiles]
                 args = [x.replace("@INPUT@", infilename).replace('@OUTPUT@', sole_output)\
                         for x in base_args]
+                args = self.replace_outputs(args, self.get_target_private_dir(target), outfilelist)
                 args = [x.replace("@SOURCE_DIR@", self.environment.get_source_dir()).replace("@BUILD_DIR@", self.get_target_private_dir(target))
                         for x in args]
-                cmdlist = [exe_file] + args
+                cmdlist = exe_arr + args
                 elem = NinjaBuildElement(outfiles, 'CUSTOM_COMMAND', infilename)
                 elem.add_item('DESC', 'Generating $out')
                 if isinstance(exe, build.BuildTarget):
@@ -758,18 +1007,85 @@ class NinjaBackend(backends.Backend):
                 elem.add_item('COMMAND', cmdlist)
                 elem.write(outfile)
 
+    def scan_fortran_module_outputs(self, target):
+        compiler = None
+        for c in self.build.compilers:
+            if c.get_language() == 'fortran':
+                compiler = c
+                break
+        if compiler is None:
+            self.fortran_deps[target.get_basename()] = {}
+            return
+        modre = re.compile(r"\s*module\s+(\w+)", re.IGNORECASE)
+        module_files = {}
+        for s in target.get_sources():
+            # FIXME, does not work for generated Fortran sources,
+            # but those are really rare. I hope.
+            if not compiler.can_compile(s):
+                continue
+            for line in open(os.path.join(self.environment.get_source_dir(), target.subdir, s)):
+                modmatch = modre.match(line)
+                if modmatch is not None:
+                    modname = modmatch.group(1)
+                    if modname.lower() == 'procedure': # MODULE PROCEDURE construct
+                        continue
+                    if modname in module_files:
+                        raise InvalidArguments('Namespace collision: module %s defined in two files %s and %s.' %
+                                               (modname, module_files[modname], s))
+                    module_files[modname] = s
+        self.fortran_deps[target.get_basename()] = module_files
+
+    def get_fortran_deps(self, compiler, src, target):
+        mod_files = []
+        usere = re.compile(r"\s*use\s+(\w+)", re.IGNORECASE)
+        dirname = os.path.join(self.get_target_dir(target), target.get_basename() + '.dir')
+        tdeps= self.fortran_deps[target.get_basename()]
+        for line in open(src):
+            usematch = usere.match(line)
+            if usematch is not None:
+                usename = usematch.group(1)
+                if usename not in tdeps:
+                    # The module is not provided by any source file. This is due to
+                    # a) missing file/typo/etc
+                    # b) using a module provided by the compiler, such as OpenMP
+                    # There's no easy way to tell which is which (that I know of)
+                    # so just ignore this and go on. Ideally we would print a
+                    # warning message to the user but this is a common occurrance,
+                    # which would lead to lots of distracting noise.
+                    continue
+                mod_source_file = tdeps[usename]
+                # Check if a source uses a module it exports itself.
+                # Potential bug if multiple targets have a file with
+                # the same name.
+                if mod_source_file == os.path.split(src)[1]:
+                    continue
+                mod_name = compiler.module_name_to_filename(usematch.group(1))
+                mod_files.append(os.path.join(dirname, mod_name))
+        return mod_files
+
     def generate_single_compile(self, target, outfile, src, is_generated=False, header_deps=[], order_deps=[]):
+        if(isinstance(src, str) and src.endswith('.h')):
+            raise RuntimeError('Fug')
+        if isinstance(src, RawFilename) and src.fname.endswith('.h'):
+            raise RuntimeError('Fug')
+        extra_orderdeps = []
         compiler = self.get_compiler_for_source(src)
         commands = self.generate_basic_compiler_args(target, compiler)
-        commands.append(compiler.get_include_arg(self.get_target_private_dir(target)))
-        if is_generated:
+        commands += compiler.get_include_args(self.get_target_private_dir(target))
+        if isinstance(src, RawFilename):
+            rel_src = src.fname
+        elif is_generated:
             if '/' in src:
                 rel_src = src
             else:
                 rel_src = os.path.join(self.get_target_private_dir(target), src)
+                abs_src = os.path.join(self.environment.get_source_dir(), rel_src)
         else:
             rel_src = os.path.join(self.build_to_src, target.get_source_subdir(), src)
-        if os.path.isabs(src):
+            abs_src = os.path.join(self.environment.get_build_dir(), rel_src)
+        if isinstance(src, RawFilename):
+            src_filename = src.fname
+        elif os.path.isabs(src):
             src_filename = os.path.basename(src)
         else:
             src_filename = src
@@ -793,27 +1109,46 @@ class NinjaBackend(backends.Backend):
             for d in i.get_incdirs():
                 expdir =  os.path.join(basedir, d)
                 fulldir = os.path.join(self.environment.get_source_dir(), expdir)
-                barg = compiler.get_include_arg(expdir)
-                sarg = compiler.get_include_arg(fulldir)
-                commands.append(barg)
-                commands.append(sarg)
+                bargs = compiler.get_include_args(expdir)
+                sargs = compiler.get_include_args(fulldir)
+                commands += bargs
+                commands += sargs
         if self.environment.coredata.use_pch:
             commands += self.get_pch_include_args(compiler, target)
         crstr = ''
         if target.is_cross:
             crstr = '_CROSS'
         compiler_name = '%s%s_COMPILER' % (compiler.get_language(), crstr)
+        extra_deps = []
+        if compiler.get_language() == 'fortran':
+            extra_deps += self.get_fortran_deps(compiler, abs_src, target)
+            # Dependency hack. Remove once multiple outputs in Ninja is fixed:
+            # https://groups.google.com/forum/#!topic/ninja-build/j-2RfBIOd_8
+            for modname, srcfile in self.fortran_deps[target.get_basename()].items():
+                modfile = os.path.join(self.get_target_dir(target), target.get_basename() + '.dir',
+                                       compiler.module_name_to_filename(modname))
+                if srcfile == src:
+                    depelem = NinjaBuildElement(modfile, 'FORTRAN_DEP_HACK', rel_obj)
+                    depelem.write(outfile)
+            commands += compiler.get_module_outdir_args(os.path.join(self.get_target_dir(target), target.get_basename() + '.dir'))
 
         element = NinjaBuildElement(rel_obj, compiler_name, rel_src)
         for d in header_deps:
-            if not '/' in d:
+            if isinstance(d, RawFilename):
+                d = d.fname
+            elif not '/' in d:
                 d = os.path.join(self.get_target_private_dir(target), d)
             element.add_dep(d)
+        for d in extra_deps:
+            element.add_dep(d)
         for d in order_deps:
-            if not '/' in d:
+            if isinstance(d, RawFilename):
+                d = d.fname
+            elif not '/' in d :
                 d = os.path.join(self.get_target_private_dir(target), d)
             element.add_orderdep(d)
         element.add_orderdep(pch_dep)
+        element.add_orderdep(extra_orderdeps)
         element.add_item('DEPFILE', dep_file)
         element.add_item('ARGS', commands)
         element.write(outfile)
@@ -830,10 +1165,10 @@ class NinjaBackend(backends.Backend):
         commands = []
         commands += self.generate_basic_compiler_args(target, compiler)
         just_name = os.path.split(header)[1]
-        commands += compiler.gen_pch_args(just_name, source, dst)
-        
+        (objname, pch_args) = compiler.gen_pch_args(just_name, source, dst)
+        commands += pch_args
         dep = dst + '.' + compiler.get_depfile_suffix()
-        return (commands, dep, dst)
+        return (commands, dep, dst, [objname])
 
     def generate_gcc_pch_command(self, target, compiler, pch):
         commands = []
@@ -841,10 +1176,11 @@ class NinjaBackend(backends.Backend):
         dst = os.path.join(self.get_target_private_dir(target),
                            os.path.split(pch)[-1] + '.' + compiler.get_pch_suffix())
         dep = dst + '.' + compiler.get_depfile_suffix()
-        return (commands, dep, dst)
+        return (commands, dep, dst, []) # Gcc does not create an object file during pch generation.
 
     def generate_pch(self, target, outfile):
         cstr = ''
+        pch_objects = []
         if target.is_cross:
             cstr = '_CROSS'
         for lang in ['c', 'cpp']:
@@ -856,12 +1192,13 @@ class NinjaBackend(backends.Backend):
             compiler = self.get_compiler_for_lang(lang)
             if compiler.id == 'msvc':
                 src = os.path.join(self.build_to_src, target.get_source_subdir(), pch[-1])
-                (commands, dep, dst) = self.generate_msvc_pch_command(target, compiler, pch)
+                (commands, dep, dst, objs) = self.generate_msvc_pch_command(target, compiler, pch)
                 extradep = os.path.join(self.build_to_src, target.get_source_subdir(), pch[0])
             else:
                 src = os.path.join(self.build_to_src, target.get_source_subdir(), pch[0])
-                (commands, dep, dst) = self.generate_gcc_pch_command(target, compiler, pch[0])
+                (commands, dep, dst, objs) = self.generate_gcc_pch_command(target, compiler, pch[0])
                 extradep = None
+            pch_objects += objs
             rulename = compiler.get_language() + cstr + '_PCH'
             elem = NinjaBuildElement(dst, rulename, src)
             if extradep is not None:
@@ -869,6 +1206,7 @@ class NinjaBackend(backends.Backend):
             elem.add_item('ARGS', commands)
             elem.add_item('DEPFILE', dep)
             elem.write(outfile)
+        return pch_objects
 
     def generate_shsym(self, outfile, target):
         target_name = self.get_target_filename(target)
@@ -879,7 +1217,7 @@ class NinjaBackend(backends.Backend):
             elem.add_item('CROSS', '--cross-host=' + self.environment.cross_info['name'])
         elem.write(outfile)
 
-    def generate_link(self, target, outfile, outname, obj_list, linker):
+    def generate_link(self, target, outfile, outname, obj_list, linker, extra_args=[]):
         if isinstance(target, build.StaticLibrary):
             linker_base = 'STATIC'
         else:
@@ -901,22 +1239,37 @@ class NinjaBackend(backends.Backend):
         elif isinstance(target, build.SharedLibrary):
             commands += linker.get_std_shared_lib_link_args()
             commands += linker.get_pic_args()
-            commands += linker.get_soname_args(target.name, abspath)
+            if hasattr(target, 'soversion'):
+                soversion = target.soversion
+            else:
+                soversion = None
+            commands += linker.get_soname_args(target.name, abspath, soversion)
         elif isinstance(target, build.StaticLibrary):
             commands += linker.get_std_link_args()
         else:
             raise RuntimeError('Unknown build target type.')
-        dependencies = target.get_dependencies()
+        # Link arguments of static libraries are not put in the command line of
+        # the library. They are instead appended to the command line where
+        # the static library is used.
+        if linker_base == 'STATIC':
+            dependencies = []
+        else:
+            dependencies = target.get_dependencies()
         commands += self.build_target_link_arguments(linker, dependencies)
         commands += target.link_args
         # External deps must be last because target link libraries may depend on them.
         if not(isinstance(target, build.StaticLibrary)):
             for dep in target.get_external_deps():
                 commands += dep.get_link_args()
+        for d in target.get_dependencies():
+            if isinstance(d, build.StaticLibrary):
+                for dep in d.get_external_deps():
+                    commands += dep.get_link_args()
         commands += linker.build_rpath_args(self.environment.get_build_dir(),\
                                             target.get_rpaths(), target.install_rpath)
         if self.environment.coredata.coverage:
             commands += linker.get_coverage_link_args()
+        commands += extra_args
         dep_targets = [self.get_dependency_filename(t) for t in dependencies]
         dep_targets += [os.path.join(self.environment.source_dir,
                                      target.subdir, t) for t in target.link_depends]
