@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-# Copyright 2012-2013 Jussi Pakkanen
+# Copyright 2012-2014 Jussi Pakkanen
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -192,6 +192,10 @@ def run_tests():
         javatests = gather_tests('test cases/java')
     else:
         javatests = []
+    if shutil.which('mcs'):
+        cstests = gather_tests('test cases/csharp')
+    else:
+        cstests = []
     if shutil.which('valac'):
         valatests = gather_tests('test cases/vala')
     else:
@@ -204,6 +208,10 @@ def run_tests():
         objctests = gather_tests('test cases/objc')
     else:
         objctests = []
+    if shutil.which('gfortran'):
+        fortrantests = gather_tests('test cases/fortran')
+    else:
+        fortrantests = []
     try:
         os.mkdir(test_build_dir)
     except OSError:
@@ -236,6 +244,11 @@ def run_tests():
         [run_and_log(logfile, t) for t in javatests]
     else:
         print('\nNot running Java tests.\n')
+    if len(cstests) > 0:
+        print('\nRunning C# tests.\n')
+        [run_and_log(logfile, t) for t in cstests]
+    else:
+        print('\nNot running C# tests.\n')
     if len(valatests) > 0:
         print('\nRunning Vala tests.\n')
         [run_and_log(logfile, t) for t in valatests]
@@ -247,14 +260,17 @@ def run_tests():
     else:
         print('\nNot running Rust tests.\n')
     if len(objctests) > 0:
-        print('\nRunning extra language tests.\n')
+        print('\nRunning Objective C tests.\n')
         [run_and_log(logfile, t) for t in objctests]
     else:
-        print('\nNo extra language tests on this platform.\n')
+        print('\nNo Objective C tests on this platform.\n')
+    if len(fortrantests) > 0:
+        print('\nRunning Fortran tests.\n')
+        [run_and_log(logfile, t) for t in fortrantests]
+    else:
+        print('\nNo Fortran tests on this platform.\n')
 
 def check_file(fname):
-    if fname.endswith('parsetab.py'): # Autogenerated
-        return True
     linenum = 1
     for line in open(fname, 'rb').readlines():
         if b'\t' in line:
